# This class is a copy of resource_api simpleprovider.
# It send the should and the is to create, update and delete functions.
class Puppet::Provider::F5SimpleProvider

  def canonicalize(context, resources)
    resources.map do |resource|
      context.type.attributes.each do |attr_name, attr|
        if attr.key?(:canonicalize)
          Puppet.debug('%{t}: canonicalize: %{object}' % { t: Time.now.to_i, object: attr_name })
          if resource.key?(attr_name)
            case attr[:canonicalize]
            when Hash
              resource[attr_name] = attr[:canonicalize].merge(resource[attr_name])
            when Array
              resource[attr_name] = attr[:canonicalize] + (resource[attr_name] || [])
            when Proc
              resource[attr_name] = attr[:canonicalize].call(resource[attr_name])
            end
          else
            case attr[:canonicalize]
            when Hash, Array
              resource[attr_name] = attr[:canonicalize]
            end
          end
        end
      end
      resource
    end
  end

  def get(context)
    # raise Puppet::DevError, 'Stack'
    # Return the cached resulsts if we have them, or query and process all of the resources
    @results ||= begin
      if top_level_resource?(context)
        # This is a top level resource, we can get all of them and process the returning list
        context.debug('Generating request and parsing top level resources')
        process_elements(context, context.transport.get(context.type.definition[:baseurl]))
      else
        # This is a contained resource, we need to get the parent resource first and then get all the children
        context.debug('Generating request and parsing contained resources')
        context.type.definition[:baseurl].map do |container_type, base_url|
          # Get all the containers of the given type and map them to the full path
          containers = context.transport.get(base_url.split('/*/')[0]).map do |container|
            if container.key?(:subPath)
              "/#{container[:partition]}/#{container[:subPath]}/#{container[:name]}"
            else
              "/#{container[:partition]}/#{container[:name]}"
            end
          end
          # Get all the children of the containers, and apply the process_element function to each of them
          context.transport.get_all(base_url, containers) do |element, container_name|
            element[:container] = container_type.to_s
            element[:containerName] = container_name
            process_element(context, element)
          end
        end.flatten(1)
      end
    end
  end

  def set(context, changes)
    changes.each do |name, change|
      # Check to see if a current state exists or if it should be retreived
      is = change[:is] if change.key?(:is)
      is ||= get(context).find {|r| compare_namevars(context.type.namevars, r, name) }
      
      context.type.check_schema(is) unless change.key?(:is)

      should = change[:should]

      raise 'F5SimpleProvider cannot be used with a Type that is not ensurable' unless context.type.ensurable?

      is = create_absent(name) if is.nil?
      should = create_absent(name) if is.nil?

      if is[:ensure].to_s == 'absent' && should[:ensure].to_s == 'present'
        context.creating(name) do
          create(context, name, should)
        end
      elsif is[:ensure].to_s == 'present' && should[:ensure].to_s == 'present'
        context.updating(name) do
          update(context, name, should)
        end
      elsif is[:ensure].to_s == 'present' && should[:ensure].to_s == 'absent'
        context.deleting(name) do
          delete(context, name, should)
        end
      end
    end
  end

  def create(context, _id, should)
    result = {}
    context.type.attributes.each do |attr_name, _attr|
      result[camelcase(attr_name).to_sym] = resource_to_json(should[attr_name]) unless filter_attr?(should, attr_name)
    end
    result[:name] = result.delete(:nametag)
    context.debug("Creating under commit #{should[:commit]}") unless should[:commit].nil?
    context.transport.post(url(context, 'create', should), result, should[:commit])
  end

  def update(context, id, should)
    result = {}
    context.type.attributes.each do |attr_name, _attr|
      result[camelcase(attr_name).to_sym] = resource_to_json(should[attr_name]) unless filter_attr?(should, attr_name)
    end
    context.debug("Creating under commit #{should[:commit]}") unless should[:commit].nil?
    context.transport.put(url(context, 'update', should), result, should[:commit])
  end

  def delete(context, id, should)
    context.debug("Creating under commit #{should[:commit]}") unless should[:commit].nil?
    context.transport.delete(url(context, 'delete', should), should[:commit])
  end

  # Determines if resource type is a top level resource or a contained resource
  # Returns Boolean
  def top_level_resource?(context)
    nil_or_empty?(context.type.definition[:attributes][:container_name])
  end

  def process_element(context, element)
    result = {}
    context.type.attributes.each do |attr_name, attr|
      result[attr_name] = case attr_name
        when :ensure then 'present'
        when :nametag then json_to_resource(element[:name], attr)
        when :sub_path then json_to_resource(element.fetch(:subPath, ''), attr)
        else json_to_resource(element[camelcase(attr_name).to_sym], attr)
      end
    end
    result = canonicalize(context, [result])[0]
    result[:title] = generate_title(result)
    Puppet.debug('%{t}: process_element %{object}' % { t: Time.now.to_i, object: result })
    return result
  end


  def process_elements(context, elements)
    context.debug("Retrive all resources for #{context.type.definition[:name]}. Total: #{elements.count}")
    elements.map { |element| process_element(context, element) }
  end

  def generate_title(resource)
    '%{container}%{container_name}%{partition}%{sub_path}%{nametag}' % {
      container: !nil_or_empty?(resource[:container]) ? "/#{resource[:container]}" : '',
      container_name: !nil_or_empty?(resource[:container_name]) ? "/#{resource[:container_name]}/" : '',
      partition: !nil_or_empty?(resource[:partition]) ? "~#{resource[:partition]}~" : '',
      sub_path: !nil_or_empty?(resource[:sub_path]) ? "#{resource[:sub_path]}~" : '',
      nametag: resource[:nametag],
    }
  end

  def nil_or_empty?(value)
    value.nil? || value.empty?
  end

  def filter_attr?(should, attr_name)
    should[attr_name].nil? || attr_name == :commit || attr_name == :ensure || attr_name == :container_name || attr_name == :container
  end

  def url(context, mode, should)
    # Determine the base url based on the mode and if the resource is a top level resource or a contained resource
    base_url = top_level_resource?(context) ?
      context.type.definition[:baseurl] : context.type.definition[:baseurl][should[:container].to_sym].gsub('*', should[:container_name])

    # return the base_url for reasource creation
    return base_url if mode == 'create'

    # return the base_url with the path to the container resource for contained resources
    return '%{base_url}/%{partition}%{sub_path}%{nametag}' % {
      base_url: base_url,
      partition: !nil_or_empty?(should[:partition]) ? "~#{should[:partition]}~" : '',
      sub_path: !nil_or_empty?(should[:sub_path]) ? "#{should[:sub_path]}~" : '',
      nametag: should[:nametag],
    }
  end

  # Recursively change keys to snakecase strings
  def json_to_resource(value,attr)
    case value
    when Hash
      h = value.reject{|k,_| attr.fetch(:struct_skip_keys, []).include?(k) }.map do |k,v|
        [snakecase(k.to_s), json_to_resource(v, attr)]
      end
      Hash[h]
    when Array
      value.map { |v| json_to_resource(v, attr) }
    else
      value
    end
  end

  # Recursively changes keys to camelcase strings
  def resource_to_json(value)
    case value
    when Hash
      h = value.map do |k,v|
        [camelcase(k.to_s), resource_to_json(v)]
      end
      Hash[h]
    when Array
      value.map { |v| resource_to_json(v) }
    else
      value
    end
  end

  def compare_namevars(namevar, resource, id)
    namevars.all? { |namevar| resource[namevar] == id[namevar] }
  end

  def create_absent(id)
    id.merge({ensure: 'absent'})
  end
  
  def camelcase(str, *separators)
    case separators.first
    when Symbol, TrueClass, FalseClass, NilClass
      first_letter = separators.shift
    end
    separators = ['_', '\s', '-'] if separators.empty?
    str = str.to_s if str.is_a?(Symbol)
    separators.each do |s|
      s = Regexp.escape(s) unless s.include? '\\'
      str = str.gsub(/(?:#{s}+)([a-zA-Z])/){ $1.upcase }
    end
    case first_letter
    when :upper, true
      str = str.gsub(/(\A|\s)([a-z])/){ $1 + $2.upcase }
    when :lower, false
      str = str.gsub(/(\A|\s)([A-Z])/){ $1 + $2.downcase }
    end
    str
  end

  def snakecase(str)
    str.
    gsub(/([A-Z]+)([A-Z][a-z])/,'\1_\2').
    gsub(/([a-z\d])([A-Z])/,'\1_\2').
    tr('-', '_').
    gsub(/\s/, '_').
    gsub(/__+/, '_').
    delete_prefix('_').
    delete_suffix('_').
    downcase
  end
end
