# frozen_string_literal: true

require 'date'
require 'etc'
require 'json'
require 'async'
require 'async/barrier'
require 'async/semaphore'
require 'async/http/client'
require 'async/http/endpoint'
require 'async/http/proxy'

module Puppet::Transport
  # The main connection class to a F5 endpoint
  class F5
    # Initialise this transport with a set of credentials
    def initialize(context, connection_info)
      # because `password` is marked sensitive, we can log it here and it will be masked
      context.debug("Connecting to #{connection_info[:username]}:#{connection_info[:password]}@#{connection_info[:host]}:#{connection_info[:port]}")
      # store the credentials for later use
      # alternatively, connect right here
      @connection_info = connection_info
    end

    # Verifies that the stored credentials are valid, and that we can talk to the target
    # TODO: Token management
    def verify(context)
      context.debug("Checking connection to #{@connection_info[:host]}:#{@connection_info[:port]}")
      !token.nil?
    end

    # Retrieve facts from the target and return in a hash
    def facts(context)
      @facts ||= parse_device_facts(fetch_device_facts(context))
    end

    def parse_device_facts(response)
      facts = {}
      facts['osfamily'] = 'f5'
      facts['hostname'] = response[:hostname]
      facts['fqdn'] = response[:hostname]
      facts
    end

    def fetch_device_facts(context)
      context.debug("Retrieving #{@connection_info[:host]} facts")
      api.get 'tm/sys/global-settings'
    end

    def token
      api.token
    end

    def get(url)
      api.get(url)
    end

    def get_all(url, containers)
      Sync do
        containers.map do |container|
          Async do
            api.get("#{url.gsub('*', container.gsub('/','~'))}").map do |response|
              yield response, container
            end
          end
        end.map(&:wait)
      end.flatten(1)
    end

    def post(url, body, commit)
      api.post(url, body, commit)
    end

    def put(url, body, commit)
      api.put(url, body, commit)
    end

    def delete(url, commit)
      api.delete(url, commit)
    end

    def commit(name)
      api.commit(name)
    end

    def changes_pending?(name)
      api.changes_pending?(name)
    end

    private

    def api
      @api ||= API.new(@connection_info)
    end

    # Implements low level calls to the REST API.
    class API
      attr_reader :host, :port, :user, :password, :ssl_verify, :proxy, :conn_limit

      RESPONSE_MESSAGES = {
        400 => '400 - Bad Request. Malformed HTTP request. Incorrect name for a resource in a request',
        401 => '401 - Unauthoried. Missing HTTP authorization header. Insufficient permissions for the credentials supplied for an administrator',
        403 => '403 - Forbidden. Insufficient permissions for the credentials supplied for an administrator. attempt to perform an unsupported action, such as deleting a property',
        404 => '404 - Not Found. Attempting to access a resource that no longer exists in the database',
        409 => '409 - Conflict. Attempting to create a resource that already exists',
        415 => '415 - Unsupported Media Type. Specifying an incorrect Content-Type header value. Specifying a malformed JSON body with a POST or PUT request',
        500 => '500 - Internal Server Error. Attempting to access iControl REST when the process is not running. Other internal errors.',
        501 => '501 - Not Implemented. Attempting to access a endpoint that does not exist. Attempting to invoke an unsupported tmsh command through iControl REST',
      }.freeze

      def initialize(connection_info)
        Puppet.debug('Initialize API %{connection_info}' % { connection_info: connection_info })
        @host = connection_info[:host] || connection_info[:address]
        @port = connection_info.key?(:port) ? connection_info[:port].to_i : 443
        @user = connection_info[:user] || connection_info[:username]
        @password = connection_info[:password].unwrap unless connection_info[:password].nil?
        @ssl_verify = connection_info[:ssl_verify].nil? ? true : connection_info[:ssl_verify]
        @proxy = connection_info[:proxy] || ENV['HTTPS_PROXY'] || ENV['https_proxy']
        @conn_limit = connection_info[:conn_limit] ? connection_info[:conn_limit].to_i : 4
        @transactions = {}

        client()
      end

      def client
        @client ||= begin
          # Disable SSL verification if ssl_verify is false
          ssl_context = OpenSSL::SSL::SSLContext.new.tap do |ctx|
            ctx.verify_mode = OpenSSL::SSL::VERIFY_NONE unless ssl_verify
          end

          # Create an endpoint to connect to host
          endpoint = Async::HTTP::Endpoint.parse("https://#{host}:#{port}", ssl_context: ssl_context)

          # Modify endpoint to be procied if proxy was provided
          if proxy
            proxy_client = Async::HTTP::Client.new(Async::HTTP::Endpoint.parse(proxy))
            proxy = Async::HTTP::Proxy.new(proxy_client, endpoint.authority(false))
            endpoint = proxy.wrap_endpoint(endpoint)
          end

          Async::HTTP::Client.new(endpoint, connection_limit: conn_limit)
        end
      end

      def fetch_token(user, password)
        Sync do |task|
          body = [JSON.dump({ username: user, password: password, loginProviderName: 'tmos' })]
          response = client.post('/mgmt/shared/authn/login', nil, body)
          handle_response(response)
          response_body = JSON.parse(response.read, symbolize_names: true)
          @token_expiration = response_body[:token][:expirationMicros]
          Puppet.debug("New Token expires in #{@token_expiration}")
          response_body[:token][:token]
        end
      end

      def token
        Puppet.debug('Retrieve token')
        @token = fetch_token(user, password) if token_expired?
        @token
      end

      # Check token expiration giving a 1 minute margin
      def token_expired?
        Puppet.debug("Current Token expiration #{token_expiration}")
        token_expiration <= (DateTime.now.strftime('%Q%3N').to_i - 60_000_000)
      end

      # Token expiration, if null then return curent time
      def token_expiration
        @token_expiration ||= 0
      end

      def transaction(name)
        @transactions[name] = fetch_transaction unless transaction?(name)
        @transactions[name]
      end

      def transaction?(name)
        @transactions.key?(name)
      end

      def fetch_transaction
        Puppet.debug("Fetching a new transaction")
        post('tm/transaction', {})[:transId]
      end

      def status_transaction(name)
        Puppet.debug("Getting transaction #{name} status")
        get("tm/transaction/#{transaction(name)}")[:state]
      end

      def commit(name)
        Puppet.debug("Commiting transation #{name}")
        patch("tm/transaction/#{transaction(name)}", { state: 'VALIDATING' })[:state]
      end
      
      def changes_pending?(name)
        case state = status_transaction(name)
        when 'STARTED' then false
        when 'UPDATING' then true
        else raise "Error: Transaction #{name} was found in state #{state} before commit."
        end
      end

      def get(url)
        Puppet.debug('GET request %{url}' % { url: url })
        Sync do |task|
          headers = [['X-F5-Auth-Token', token]]
          response = client.get("/mgmt/#{url}", headers)
          handle_response(response)
          parse_response(JSON.parse(response.read, symbolize_names: true))
        end
      end

      def post(url, body, commit = nil)
        Puppet.debug('POST request %{url} - %{body}' % { url: url, body: body })
        Sync do |task|
          headers = [['X-F5-Auth-Token', token]]
          headers.push(['X-F5-REST-Coordination-Id', transaction(commit).to_s]) unless commit.nil?
          response = client.post("/mgmt/#{url}", headers, [JSON.dump(body)])
          handle_response(response)
          parse_response(JSON.parse(response.read, symbolize_names: true))
        end
      end

      def put(url, body, commit = nil)
        Puppet.debug('PUT request %{url}' % { url: url })
        Sync do |task|
          headers = [['X-F5-Auth-Token', token]]
          headers.push(['X-F5-REST-Coordination-Id', transaction(commit).to_s]) unless commit.nil?
          response = client.put("/mgmt/#{url}", headers, [JSON.dump(body)])
          handle_response(response)
          parse_response(JSON.parse(response.read, symbolize_names: true))
        end
      end

      def patch(url, body, commit = nil)
        Puppet.debug('PATCH request %{url}' % { url: url })
        Sync do |task|
          headers = [['X-F5-Auth-Token', token]]
          headers.push(['X-F5-REST-Coordination-Id', transaction(commit).to_s]) unless commit.nil?
          response = client.patch("/mgmt/#{url}", headers, [JSON.dump(body)])
          handle_response(response)
          parse_response(JSON.parse(response.read, symbolize_names: true))
        end
      end

      def delete(url, commit = nil)
        Puppet.debug('DELETE request %{url}' % { url: url })
        Sync do |task|
          headers = [['X-F5-Auth-Token', token]]
          headers.push(['X-F5-REST-Coordination-Id', transaction(commit).to_s]) unless commit.nil?
          response = client.delete("/mgmt/#{url}", headers)
          handle_response(response)
        end
      end

      def handle_response(response)
        Puppet.debug('Verifying response')
        raise Puppet::ResourceError, "Error: #{response_code_message(response.status, response.body)}" unless response.success?
      end

      def response_code_message(status, body)
        return "#{body[:code]} - #{body[:message]}. APIerror: #{body.fetch(:apiError, 'N/A')}" if body.is_a?(Hash) && body.key?(:code) && body.key?(:message)
        RESPONSE_MESSAGES.fetch(status, "#{status} - Undefined status code. #{body}")
      end

      def parse_response(body)
        Puppet.debug('Parsing response')
        return body[:items] if body.key?(:items)
        body
      end
    end
  end
end
