# Shared resources cross types
class Puppet::Type::F5
  def self.base_definition
    {
      title_patterns: self.base_title_patterns,
      features: self.base_features,
      attributes: self.base_attributes,
      autobefore: self.base_autobefore,
    }
  end

  def self.base_attributes
    {
      ensure: {
        type:    'Enum[present, absent]',
        desc:    'Whether this resource should be present or absent on the target system.',
        default: 'present',
      },
      nametag: {
        type:      'String',
        desc:      'The name of the resource to manage.',
        behaviour: :namevar,
      },
      partition: {
        type:      'String',
        desc:      'Patition where the resource is created.',
        behaviour: :namevar,
        default:   'Common',
      },
      sub_path: {
        type:         'String',
        desc:         'Path inside the partition where the resource is stored.',
        behaviour:    :namevar,
        default:      '',
        canonicalize: ->(value) { value.gsub('/','~') },
      },
      full_path: {
        type:      'String', # TODO: Stdlib::Unixpath
        desc:      'The fullpath to the resource to manage. Main resource indetifier.',
        behaviour: :read_only,
      },
      kind: {
        type:      'String',
        desc:      'Unique type indetifier for F5.',
        behaviour: :read_only,
      },
      generation: {
        type:      'Integer',
        desc:      'A generation number for a resource. Modification of a resource, or a related resource, changes the value.',
        behaviour: :read_only,
      },
      self_link: {
        type: 'Variant[Pattern[/(?i:\Ahttps?:\/\/.*\z)/], Pattern[/(?i:\Ahttps:\/\/.*\z)/]]', # TODO: Variant[Stdlib::HTTPUrl, Stdlib::HTTPSUrl]
        desc: 'A link to this resource.',
        behaviour: :read_only,
      },
      commit: {
        type:      'Optional[String]',
        desc:      'Attaches this resource to this transaction. Make sure to include an f5_commit resource with the same name in the catalog.',
        behaviour: :parameter,
      },
      description: {
        type: 'Optional[String]',
        desc: 'User defined description.',
      },
    }
  end

  def self.base_autobefore
    {
      f5_commit: '$commit',
    }
  end

  def self.base_title_patterns
    [
      {
        pattern: %r{^(~|\/)(?<partition>[^~\/]+)(~|\/)(?<sub_path>[^~\/]+)(~|\/)(?<nametag>.+)$},
        desc: 'Where the partition and the name are provided by a slash separator and the resource is in a subPath',
      },
      {
        pattern: %r{^(~|\/)(?<partition>[^~\/]+)(~|\/)(?<nametag>.+)(?<sub_path>)$},
        desc: 'Where the partition and the name are provided by a slash separator',
      }
    ]
  end

  def self.base_features
    ['remote_resource', 'canonicalize']
  end

  def self.definition(overrides = {})
    self.deep_merge(self.base_definition, overrides)
  end

  def self.deep_merge(defaults, override, &block)
    self.deep_merge!(defaults.dup, override, &block)
  end

  def self.deep_merge!(defaults, override, &block)
    defaults.merge!(override) do |key, this_val, other_val|
      if this_val.is_a?(Hash) && other_val.is_a?(Hash)
        self.deep_merge(this_val, other_val, &block)
      elsif block_given?
        block.call(key, this_val, other_val)
      else
        other_val
      end
    end
  end
end