# frozen_string_literal: true

require 'puppet/resource_api'
require 'puppet/type/f5'

Puppet::ResourceApi.register_type(Puppet::Type::F5.definition({
  name: 'f5_pool',
  docs: <<-EOS,
@summary An IControlRest type for LTM pools
@example
f5_pool { 'foo':
  ensure => 'present',
}

This type provides Puppet with the capabilities to manage LTM pools on F5 systems
EOS
  baseurl: 'tm/ltm/pool',
  attributes: {
    # Map properties of https://clouddocs.f5.com/api/icontrol-rest/APIRef_tm_ltm_pool.html to Puppet attributes
    allow_nat: {
      type: 'Optional[Enum["yes", "no"]]',
      desc: 'Specifies whether the pool allows NAT to occur for connections to the pool members.',
      default: 'yes',
    },
    allow_snat: {
      type: 'Optional[Enum["yes", "no"]]',
      desc: 'Specifies whether the pool allows SNAT to occur for connections to the pool members.',
      default: 'yes',
    },
    autoscale_group_id: {
      type: 'Optional[String]',
      desc: 'autoscale-group id to which pool members belong to.',
    },
    gateway_failsafe_device: {
      type: 'Optional[String]',
      desc: 'Specifies that the pool is a gateway failsafe pool in a redundant configuration. This string identifies the device that will failover when the monitor reports the pool member down. By default the device string is empty.',
    },
    ignore_persisted_weight: {
      type: 'Optional[Enum["disabled", "enabled"]]',
      desc: 'Do not count the weight of persisted connections on pool members when making load balancing decisions.',
      default: 'disabled',
    },
    ip_tos_to_client: {
      type: 'Optional[String]',
      desc: 'Specifies the Type of Service (ToS) level to use when sending packets to a client. 65534 (mimic) specifies that the system sets the ToS level of outgoing packets to the same ToS level of the most-recently received incoming packet. The default value is 65535 (pass-through).',
      default: 'pass-through',
    },
    ip_tos_to_server: {
      type: 'Optional[String]',
      desc: '	Specifies the Type of Service (ToS) level to use when sending packets to a server. 65534 (mimic) specifies that the system sets the ToS level of outgoing packets to the same ToS level of the most-recently received incoming packet. The default value is 65535 (pass-through).',
      default: 'pass-through',
    },
    link_qos_to_client: {
      type: 'Optional[String]',
      desc: '	Specifies the Quality of Service (QoS) level to use when sending packets to a client. The default value is 65535 (pass-through).',
      default: 'pass-through',
    },
    link_qos_to_server: {
      type: 'Optional[String]',
      desc: 'Specifies the Quality of Service (QoS) level to use when sending packets to a server. The default value is 65535 (pass-through).',
      default: 'pass-through',
    },
    load_balancing_mode: {
      type: 'Optional[Enum["dynamic-ratio-member", "dynamic-ratio-node", "fastest-app-response", "fastest-node", "least-connections-member", "least-connections-node", "least-sessions", "observed-member", "observed-node", "predictive-member", "predictive-node", "ratio-least-connections-member", "ratio-least-connections-node", "ratio-member", "ratio-node", "ratio-session", "round-robin", "weighted-least-connections-member", "weighted-least-connections-node"]]',
      desc: 'Specifies the modes that the system uses to load balance name resolution requests among the members of this pool. See “help pool” for a description of each loading balancing mode.',
      default: 'round-robin',
    },
    min_active_members: {
      type: 'Optional[Integer]',
      desc: 'Specifies the minimum number of members that must be up for traffic to be confined to a priority group when using priority-based activation. The default value is 0 (zero). An active member is a member that is up (not marked down) and is handling fewer connections than its connection limit.',
      default: 0,
    },
    min_up_members: {
      type: 'Optional[Integer]',
      desc: 'Specifies the minimum number of pool members that must be up; otherwise, the system takes the action specified in the min-up-members-action option. Use this option for gateway pools in a redundant system where a unit number is applied to a pool. This indicates that the pool is only configured on the specified unit.',
      default: 0,
    },
    min_up_members_action: {
      type: 'Optional[Enum["failover", "failover-restart", "none"]]',
      desc: 'Specifies the action to take if the min-up-members-checking is enabled and the number of active pool members falls below the number specified in min-up-members. The default value is failover.',
      default: 'failover',
    },
    min_up_members_checking: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Enables or disables the min-up-members feature. If you enable this feature, you must also specify a value for both the min-up-members and min-up-members-action options.',
      default: 'disabled',
    },
    monitor: {
      type: 'Optional[String]',
      desc: 'Specifies the health monitors that the system uses to determine whether it can use this pool for load balancing. The monitor marks the pool up or down based on whether the monitor(s) are successful. You can specify a single monitor, multiple monitors “http and https”, or a “min” rule, “min 1 of { http https }”. You may remove the monitor by specifying “none”.',
    },
    queue_depth_limit: {
      type: 'Optional[Integer]',
      desc: 'Specifies the maximum number of connections that may simultaneously be queued to go to any member of this pool. The default is zero which indicates there is no limit.',
      default: 0,
    },
    queue_on_connection_limit: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Enable or disable queuing connections when pool member or node connection limits are reached. When queuing is not enabled, new connections are reset when connection limits are met.',
      default: 'disabled',
    },
    queue_time_limit: {
      type: 'Optional[Integer]',
      desc: 'Specifies the maximum time, in milliseconds, a connection will remain enqueued. The default is zero which indicates there is no limit.',
      default: 0,
    },
    reselect_tries: {
      type: 'Optional[Integer]',
      desc: 'Specifies the number of times the system tries to contact a pool member after a passive failure. A passive failure consists of a server-connect failure or a failure to receive a data response within a user-specified interval. The default is 0 (zero), which indicates no reselect attempts.',
      default: 0,
    },
    service_down_action: {
      type: 'Optional[Enum["none", "reset", "drop", "reselect"]]',
      desc: 'Specifies the action to take if the service specified in the pool is marked down. The default value is none.',
      default: 'none',
    },
    slow_ramp_time: {
      type: 'Optional[Integer]',
      desc: '	Specifies, in seconds, the ramp time for the pool. This provides the ability to cause a pool member that has just been enabled, or marked up, to receive proportionally less traffic than other members in the pool. The proportion of traffic the member accepts is determined by how long the member has been up in comparison to the slow-ramp-time setting for the pool.For example, if the load-balancing-mode of a pool is round-robin, and it has a slow-ramp-time of 60 seconds, when a pool member has been up for only 30 seconds, the pool member receives approximately half the amount of new traffic as other pool members that have been up for more than 60 seconds. After the pool member has been up for 45 seconds, it receives approximately three quarters of the new traffic.The slow ramp time is particularly useful when used with the least-connections-member load balancing mode. The default value is 10.',
      default: 10,
    },
  },
}))
