# frozen_string_literal: true

require 'puppet/resource_api'
require 'puppet/type/f5'

f5_pool_member_def = Puppet::Type::F5.definition({
  name: 'f5_pool_member',
  docs: <<-EOS,
@summary a f5_pool_member type
@example
f5_pool_member { 'foo':
  ensure => 'present',
}

This type provides Puppet with the capabilities to manage LTM pool members on F5 systems
EOS
  baseurl: {
    pool: 'tm/ltm/pool/*/members'
  },
  contained: true,
  title_patterns: [
    {
      pattern: %r{^\/(?<container>(pool)+)\/(?<container_name>(?>(~|\/)[^~\/]+){2,})\/(~|\/)(?<partition>[^~\/]+)(~|\/)(?<sub_path>[^~\/]+)(~|\/)(?<nametag>.+)$},
      desc: 'Where the full name is provided, this include the container, the container_name, partition, sub_path and the nametag of the member.',
    },
    {
      pattern: %r{^\/(?<container>(pool)+)\/(?<container_name>(?>(~|\/)[^~\/]+){2,})\/(~|\/)(?<partition>[^~]+)(~|\/)(?<nametag>.+)(?<sub_path>)$},
      desc: 'Where the full name is provided, this include the container, the container_name, partition and the nametag of the member.',
    },
  ],
  attributes: {
    container_name: {
      type:         'Variant[Pattern[/\A(?>\/[^\/]+){2,}\Z/],Pattern[/\A(?>~[^~]+){2,}\Z/]]',
      desc:         'Partition and name of the container where the member exists.',
      behaviour:    :namevar,
      canonicalize: ->(value) { value.gsub('/','~') },
    },
    container: {
      type: 'Enum[pool]',
      desc: 'Type of the container where the member exists. Can be pool.',
      behaviour: :namevar,
      default: 'pool',
    },
    # Map properties of https://clouddocs.f5.com/api/icontrol-rest/APIRef_tm_ltm_pool_members.html to Puppet attributes
    address: {
      type: 'Optional[String]',
      desc: 'IP address of a pool member if a node by the given name does not already exist.',
      behaviour: :read_only,
    },
    connection_limit: {
      type: 'Optional[Integer]',
      desc: 'Specifies the maximum number of concurrent connections allowed for a pool member. The default value is 0 (zero).',
      default: 0,
    },
    dynamic_ratio: {
      type: 'Optional[Integer]',
      desc: 'Specifies a range of numbers that you want the system to use in conjunction with the ratio load balancing method. The default value is 1.',
      default: 1,
    },
    ephemeral: {
      type: 'Optional[Enum["false", "true"]]',
      desc: 'Specifies whether the pool member is an ephemeral member. The default value is false.',
      behaviour: :read_only,
      default: 'false',
    },
    inherit_profile: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Specifies whether the pool member inherits the encapsulation profile from the parent pool. The default value is enabled. If you disable inheritance, no encapsulation takes place, unless you specify another encapsulation profile for the pool member using the profiles attribute.',
      default: 'enabled',
    },
    logging: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Specifies whether the pool member\'s monitor(s) actions will be logged. Logs are stored in /var/log/monitors/',
      default: 'disabled',
    },
    monitor: {
      type: 'Optional[String]',
      desc: 'Displays the health monitors that are configured to monitor the pool member, and the status of each monitor. The default value is default.',
      default: 'default',
    },
    priority_group: {
      type: 'Optional[Integer]',
      desc: 'Specifies the priority group within the pool for this pool member. The priority group number specifies that traffic is directed to that member before being directed to a member of a lower priority. The default value is 0.',
      default: 0,
    },
    rate_limit: {
      type: 'Optional[Variant[Integer,String]]',
      desc: 'Specifies the maximum number of connections per second allowed for a pool member. The default value is \'disabled\'.',
      default: 'disabled',
    },
    ratio: {
      type: 'Optional[Integer]',
      desc: 'Specifies the ratio weight that you want to assign to the pool member. The default value is 1.',
      default: 1,
    },
    # session: {
    #   type: 'Optional[Enum["user-enabled", "user-disabled", "monitor-enabled", "monitor-disabled"]]',
    #   desc: '	Enables or disables the pool member for new sessions. The default value is user-enabled.',
    #   default: 'user-enabled',
    # },
    # state: {
    #   type: 'Optional[Enum["user-up", "user-down", "monitor-up", "monitor-down", "forced-down", "down", "up", "fqdn-up", "fqdn-up-no-addr", "unchecked"]]',
    #   desc: 'user-down forces the pool member offline, overriding monitors. user-up reverts the user-down. When user-up, this displays the monitor state.',
    #   default: 'user-up',
    # },
    # status: {
    #   type: 'Optional[Enum["enabled", "disabled", "forced_offline"]]',
    #   desc: 'Metaparameter that simulates the experience on the UI. This controls state and session parameters.',
    #   default: 'enabled',
    #   behaviour: :parameter,
    # }
    fqdn: {
      type: <<-EOS,
        Optional[Struct[{
          Optional[autopopulate] => Enum['enabled', 'disabled'],
          Optional[tm_name]      => String,
        }]]
      EOS
      desc: 'Specifies the fully qualified domain name (FQDN) of the pool member.',
    },
  },
})

klass = Puppet::ResourceApi.register_type(f5_pool_member_def)

require 'puppet_x/f5/comparer'

# Autorequire the container
klass.autobefore(:f5_pool) do
  autos = []
  # Refresh the state of the resource
  refresh_current_state unless @rsapi_current_state
  # Obtain resource information
  container = @rsapi_current_state[:container]
  container_name = @rsapi_current_state[:container_name]
  # If we are purging, this resource needs to be delete before referenced objects
  autos << container_name.gsub(%r{\/([^\/]*?)$},'//\1') if purging? && container == 'pool'
  autos
end

# Sort fqdn array before comparison
class Puppet::Type::F5_pool_member::Fqdn
  def property_matches?(current, desired)
    PuppetX::F5::Comparer.compare(current, desired)
  end
end
