require 'puppet'
require 'puppet/parameter/boolean'

Puppet::Type.newtype(:f5_resources) do
  @doc = "This is a metatype that can manage other f5 resource types.  Any
    metaparams specified here will be passed on to any generated resources,
    so you can purge unmanaged resources but set `noop` to true so the
    purging is only logged and does not actually happen."


  newparam(:name) do
    desc "The name of the type to be managed."

    validate do |name|
      raise ArgumentError, _("Could not find resource type '%{name}'") % { name: name } unless Puppet::Type.type(name)
      raise ArgumentError, _("Can only specify f5_* types, %{name} given") % { name: name } unless /^f5_.+$/.match?(name)
    end

    munge { |v| v.to_s }
  end

  newparam(:purge, :boolean => true, :parent => Puppet::Parameter::Boolean) do
    desc "Whether to purge unmanaged resources.  When set to `true`, this will
      delete any resource that is not specified in your configuration and is not
      autorequired by any managed resources. **Note:** The `ssh_authorized_key`
      resource type can't be purged this way; instead, see the `purge_ssh_keys`
      attribute of the `user` type."

    defaultto :false

    validate do |value|
      if munge(value)
        unless @resource.resource_type.respond_to?(:instances)
          raise ArgumentError, _("Purging resources of type %{res_type} is not supported, since they cannot be queried from the system") % { res_type: @resource[:name] }
        end
        raise ArgumentError, _("Purging is only supported on types that accept 'ensure'") unless @resource.resource_type.validproperty?(:ensure)
      end
    end
  end

  newparam(:commit) do
    desc "Name of the commit that will be used for queuing the purge action"
  end

  [ :unless, :if ]. each do |param_name|
    newparam(param_name, :array_matching => :all) do

      desc(<<-EOT)
      Purge resources #{param_name.to_s} they meet the criteria.
      Criteria is defined as an array of "parameter", "operator", and "value".
      Eg:
         #{param_name.to_s} => [ 'name', '==', 'root' ]
      Operators can support "!=","==","=~",">","<","<=",">=" as an argument
      Value can be a string, integer or regex (without the enclosing slashes)
      Multiple criterias can be nested in an array, eg:
         #{param_name.to_s} => [
           [ 'name', '==', 'root' ], [ 'name', '=~', 'admin.*' ]
         ]
      EOT



      validate do |cond|
        raise ArgumentError, "must be an array" unless cond.is_a?(Array)
      end

      munge do |cond|
        if cond[0].is_a?(Array)
          cond
        else
          [ cond ]
        end
      end

      validate do |cond_param|
        case cond_param[0]
        when String
          cond = [ cond_param ]
        when Array
          cond = cond_param
        end

        cond.each do |param, operator, value|
          unless ["!=","==","=~",">","<","<=",">="].include?(operator)
            raise ArgumentError, "invalid operator #{operator}"
          end

          unless param && operator && value
            raise ArgumentError, "not enough parameters given for filter"
          end
        end
      end
    end
  end

  def check(resource)
    @checkmethod ||= "#{self[:name]}_check"
    @hascheck ||= respond_to?(@checkmethod)
    if @hascheck
      return send(@checkmethod, resource)
    else
      return true
    end
  end

  def able_to_ensure_absent?(resource)
      resource[:ensure] = :absent
  rescue ArgumentError, Puppet::Error
      err _("The 'ensure' attribute on %{name} resources does not accept 'absent' as a value") % { name: self[:name] }
      false
  end

  # Generate any new resources we need to manage.  This is pretty hackish
  # right now, because it only supports purging.
  def generate
    return [] unless self.purge?
    resource_type.instances.
      reject { |r| catalog.resource_refs.include? r.ref }.
      select { |r| check(r) }.
      select { |r| r.class.validproperty?(:ensure) }.
      select { |r| able_to_ensure_absent?(r) }.
      select { |r| conditional_purge?(r) }.
      each { |resource|
        @parameters.each do |name, param|
          resource[name] = param.value if param.metaparam?
          resource[name] = param.value if param.to_s == 'commit'
        end

        # Mark that we're purging, so transactions can handle relationships
        # correctly
        resource.purging
      }
  end

  def resource_type
    unless defined?(@resource_type)
      unless type = Puppet::Type.type(self[:name])
        raise Puppet::DevError, _("Could not find resource type")
      end
      @resource_type = type
    end
    @resource_type
  end

  # conditional_purge will evaluate the conditions given in if/unless
  def conditional_purge?(res_type)

    res = res_type.to_resource.to_hash

    if self[:unless]
      return false unless evaluate_resource(res, self[:unless])
    end

    if self[:if]
      return false if evaluate_resource(res, self[:if])
    end

    return true
  end

  def evaluate_resource(res,condition)
    condition.select  {  |param, operator, value_attr|
      Array(value_attr).select { |value|
        Array(res[param.to_sym]).select { |p|
          case operator
          when "!=", "=="
            p.to_s.send(operator, value)
          when "=~"
            p =~ Regexp.new(value)
          when ">=", "<=", ">", "<"
            case p
            when Hash, Array
              p.length.send(operator, value.to_i)
            when String, Numeric
              p.to_i.send(operator, value.to_i)
            else
              raise "Unsupported operator for #{param.to_sym}!"
            end
          end
        }.length > 0
      }.length > 0
    }.length == 0
  end
end