# frozen_string_literal: true

require 'puppet/resource_api'
require 'puppet/type/f5'

f5_rule_def = Puppet::Type::F5.definition({
  name: 'f5_rule',
  docs: <<-EOS,
@summary a f5_rule type
@example
f5_rule { 'foo':
  ensure => 'present',
}

This type provides Puppet with the capabilities to manage ...
EOS
  baseurl: {
    policy: 'tm/security/firewall/policy/*/rules',
    rulelist: 'tm/security/firewall/rule-list/*/rules',
  },
  contained: true,
  title_patterns: [
    {
      pattern: %r{^\/(?<container>(rulelist|policy)+)(?<container_name>(?>\/[^\/]+){2,})\/(?<nametag>[^\/]+)$},
      desc: 'DEPRECATED: Where the full name is provided, this include the container, the container_name and the nametag of the rule.',
    },
    {
      pattern: %r{^\/(?<container>(rulelist|policy)+)\/(?<container_name>(?>(~|\/)[^~\/]+){2,})\/(?<nametag>.+)$},
      desc: 'Where the full name is provided, this include the container, the container_name and the nametag of the rule.',
    },
  ],
  attributes: {
    container_name: {
      type: 'Variant[Pattern[/\A(?>\/[^\/]+){2,}\Z/],Pattern[/\A(?>~[^~]+){2,}\Z/]]',
      desc: 'Partition and name of the container where the rule exists.',
      behaviour: :namevar,
      default: '/Common/Global_Allow',
      canonicalize: ->(value) { value.gsub('/','~') },
    },
    container: {
      type: 'Enum[rulelist,policy]',
      desc: 'Type of the container where the rule exists. Can be either rulelist or policy.',
      behaviour: :namevar,
      default: 'policy',
    },
    action: {
      type: 'Enum[accept,reject,drop,accept-decisively,none]',
      desc: 'Specifies the action that the system takes when a rule is matched.',
    },
    ip_protocol: {
      type: 'Optional[Enum[any,tcp,udp,icmp,icmpv6,any,ospf,pim,igmp,ip,ipv6-icmp,ipv4]]',
      desc: 'Specifies the IP protocol against which the packet will be compared.',
    },
    irule: {
      type: 'Optional[String]',
      desc: 'Specifies the name of the iRule which will be triggered if the packet matches this firewall rule.',
    },
    irule_sample_rate: {
      type: 'Optional[Integer]',
      desc: 'Specifies the rate at which iRule will be triggered if the packet matches this firewall rule.',
    },
    log: {
      type: 'Optional[Enum[yes,no]]',
      desc: 'Specifies whether the packet will be logged if it matches the rule. Logging must also be enabled in the security log profile global-network configuration. Note that the statistics counter is always incremented when a packet matches a rule.',
    },
    #schedule: {
    #  type: 'Optional[String[1]]',
    #  desc: 'Specifies a schedule for the rule. See security firewall schedule. If the rule refers to a rule-list the rule-list will be enabled according to the schedule. When the rule list is enabled, the schedules defined within the rule-list will be honored.',
    #},
    service_policy: {
      type: 'Optional[String[1]]',
      desc: 'Specifies the service policy to use.',
    },
    status: {
      type: 'Optional[Enum[enabled,disabled,scheduled]]',
      desc: 'Specifies whether the rule is enabled, disabled or scheduled. A rule that is enabled is always checked. A rule that is disabled is never checked. A rule that is scheduled is checked according to the corresponding schedule configuration. A rule that is scheduled must have an associated schedule configuration.',
    },
    virtual_server: {
      type: 'Optional[String[1]]',
      desc: '',
    },
    rule_list: {
      type: 'Optional[String]',
      desc: 'Specifies a list of rules to evaluate. See security firewall rule-list. If a rule-list is specified then only the schedule and status properties effect the rule.'
    },
    destination: {
      type: <<-EOS,
        Optional[Struct[{
          Optional[addresses]     => Array[Struct[{ NotUndef[name] => String }]],
          Optional[geo]           => Array[Struct[{ NotUndef[name] => String }]],
          Optional[ports]         => Array[Struct[{ NotUndef[name] => String }]],
          Optional[address_lists] => Array[String],
          Optional[ipi_category]  => Array[String],
          Optional[port_lists]    => Array[String],
          Optional[zones]         => Array[Struct[{
            NotUndef[name]      => String,
            NotUndef[partition] => String,
            Optional[sub_path]  => String[1],
          }]],
        }]]
      EOS
      desc: 'List of all destinations.',
      struct_skip_keys: [:addressListsReference, :ipiCategoryReference, :portListsReference, :nameReference],
      canonicalize: {
        'addresses'     => [],
        'geo'           => [],
        'ports'         => [],
        'address_lists' => [],
        'ipi_category'  => [],
        'port_lists'    => [],
        'zones'         => [],
      },
    },
    source: {
      type: <<-EOS,
        Optional[Struct[{
          Optional[addresses]     => Array[Struct[{ NotUndef[name] => String }]],
          Optional[geo]           => Array[Struct[{ NotUndef[name] => String }]],
          Optional[ports]         => Array[Struct[{ NotUndef[name] => String }]],
          Optional[address_lists] => Array[String],
          Optional[port_lists]    => Array[String],
          Optional[vlans]         => Array[String],
          Optional[ipi_category]   => Array[String],
          identity                => Struct[{
            Optional[subscriber_groups] => Array[Struct[{ NotUndef[name] => String }]],
            Optional[subscribers]       => Array[Struct[{ NotUndef[name] => String }]],
          }],
          Optional[zones]         => Array[Struct[{
            NotUndef[name]      => String,
            NotUndef[partition] => String,
            Optional[sub_path]  => String[1],
          }]],
        }]]
      EOS
      desc: 'List of all sources.',
      struct_skip_keys: [:addressListsReference, :ipiCategoryReference, :portListsReference, :nameReference, :vlansReference],
      canonicalize: {
        'addresses'     => [],
        'geo'           => [],
        'ports'         => [],
        'address_lists' => [],
        'port_lists'    => [],
        'vlans'         => [],
        'ipi_category'  => [],
        'identity'      => {},
        'zones'         => [],
      },
    },
    icmp: {
      type: 'Optional[Array[Struct[{ NotUndef[name] => String[1] }]]]',
      desc: 'Specifies type and codes pairs for ICMP messages.',
      canonicalize: [],
    },
    place_after: {
      type: 'Optional[Enum[first,last]]',
      desc: 'Specifies that a new rule should be placed after first or last rule. Only useful upon resource creation.',
      default: nil,
      behaviour: :parameter,
    },
    place_before: {
      type: 'Optional[Enum[first,last]]',
      desc: 'Specifies that a new rule should be placed before first or last rule. Only useful upon resource creation.',
      default: nil,
      behaviour: :parameter,
    },
    rule_number: {
      type: 'Optional[String]',
      desc: 'Current position on the list.',
      behaviour: :read_only,
    },
    uuid: {
      type: 'Optional[String]',
      desc: 'Autogenerated rule identifier.',
      behaviour: :read_only,
    },
  },
})
f5_rule_def[:attributes].delete(:partition)
f5_rule_def[:attributes].delete(:sub_path)

klass = Puppet::ResourceApi.register_type(f5_rule_def)

require 'puppet_x/f5/comparer'

klass.autobefore(:f5_addresslist) do
  autos = []
  # Refresh the state of the resource
  refresh_current_state unless @rsapi_current_state
  # If we are purging, this resource needs to be deleted before referenced resources
  if purging?
    # gather all address_lists
    source_address_lists = @rsapi_current_state.dig(:source, 'address_lists') || []
    destination_address_lists = @rsapi_current_state.dig(:destination, 'address_lists') || []
    address_lists = source_address_lists + destination_address_lists
    # Add to the list of autos all of this objects
    autos = address_lists.map do |resource_ref|
      backslash = resource_ref.scan(%r{(?=\/)}).count
      # Add / string does not contain the sub_path
      if backslash == 2
        resource_ref.gsub(%r{\/([^\/]*?)$},'//\1')
      elsif backslash > 2
        resource_ref
      else
        raise Puppet::DevError, _("Resource ref #{resource_ref} is not a valid reference on F5 objects")
      end
    end
  end
  autos
end

klass.autobefore(:f5_portlist) do
  autos = []
  # Refresh the state of the resource
  refresh_current_state unless @rsapi_current_state
  # If we are purging, this resource needs to be deleted before referenced resources
  if purging?
    # gather all port_lists
    source_port_lists = @rsapi_current_state.dig(:source, 'port_lists') || []
    destination_port_lists = @rsapi_current_state.dig(:destination, 'port_lists') || []
    port_lists = source_port_lists + destination_port_lists
    # Add to the list of autos all of this objects
    autos = port_lists.map do |resource_ref|
      backslash = resource_ref.scan(%r{(?=\/)}).count
      # Add / string does not contain the sub_path
      if backslash == 2
        resource_ref.gsub(%r{\/([^\/]*?)$},'//\1')
      elsif backslash > 2
        resource_ref
      else
        raise Puppet::DevError, _("Resource ref #{resource_ref} is not a valid reference on F5 objects")
      end
    end
  end
  #puts "Autobefore #{title} #{purging?} #{port_lists}"
  autos
end

klass.autobefore(:f5_rulelist) do
  autos = []
  # Refresh the state of the resource
  refresh_current_state unless @rsapi_current_state
  # Obtain resource information
  container = @rsapi_current_state[:container]
  container_name = @rsapi_current_state[:container_name]
  # If we are purging, this resource needs to be delete before referenced objects
  autos << container_name.gsub(%r{\/([^\/]*?)$},'//\1') if purging? && container == 'rulelist'
  autos
end

klass.autobefore(:f5_policy) do
  autos = []
  # Refresh the state of the resource
  refresh_current_state unless @rsapi_current_state
  # Obtain resource information
  container = @rsapi_current_state[:container]
  container_name = @rsapi_current_state[:container_name]
  # If we are purging, this resource needs to be delete before referenced objects
  autos << container_name.gsub(%r{\/([^\/]*?)$},'//\1') if purging? && container == 'policy'
  autos
end

class Puppet::Type::F5_rule::Destination
  def property_matches?(current, desired)
    PuppetX::F5::Comparer.compare(current['addresses'], desired['addresses']) &&\
    PuppetX::F5::Comparer.compare(current['zones'], desired['zones']) &&\
    PuppetX::F5::Comparer.compare(current['ports'], desired['ports']) &&\
    PuppetX::F5::Comparer.compare(current['geo'], desired['geo']) &&\
    current['address_lists'].sort == desired['address_lists'].sort &&\
    current['ipi_category'].sort == desired['ipi_category'].sort &&\
    current['port_lists'].sort == desired['port_lists'].sort
  end
end

class Puppet::Type::F5_rule::Source
  def property_matches?(current, desired)
    c_subscriber_groups = current['identity'].fetch('subscriber_groups', [])
    d_subscriber_groups = desired['identity'].fetch('subscriber_groups', [])
    c_subscribers = current['identity'].fetch('subscribers', [])
    d_subscribers = desired['identity'].fetch('subscribers', [])

    PuppetX::F5::Comparer.compare(current['addresses'], desired['addresses']) &&\
    PuppetX::F5::Comparer.compare(c_subscriber_groups, d_subscriber_groups) &&\
    PuppetX::F5::Comparer.compare(current['zones'], desired['zones']) &&\
    PuppetX::F5::Comparer.compare(current['ports'], desired['ports']) &&\
    PuppetX::F5::Comparer.compare(current['geo'], desired['geo']) &&\
    PuppetX::F5::Comparer.compare(c_subscribers, d_subscribers) &&\
    current['address_lists'].sort == desired['address_lists'].sort &&\
    current['ipi_category'].sort == desired['ipi_category'].sort &&\
    current['port_lists'].sort == desired['port_lists'].sort &&\
    current['vlans'].sort == desired['vlans'].sort
  end
end

class Puppet::Type::F5_rule::Icmp
  def property_matches?(current, desired)
    PuppetX::F5::Comparer.compare(current, desired)
  end
end
