# frozen_string_literal: true

require 'puppet/resource_api'
require 'puppet/type/f5'

Puppet::ResourceApi.register_type(Puppet::Type::F5.definition({
  name: 'f5_virtual_server',
  docs: <<-EOS,
@summary An IControlRest type for LTM virtual server
@example
f5_virtual { 'foo':
  ensure => 'present',
}

This type provides Puppet with the capabilities to manage LTM virtual server on F5 systems
EOS
  baseurl: 'tm/ltm/virtual',
  attributes: {
    # Map properties of https://clouddocs.f5.com/api/icontrol-rest/APIRef_tm_ltm_virtual.html to Puppet attributes
    address_status: {
      type: 'Optional[Enum["yes", "no"]]',
      desc: 'Specifies whether the virtual will contribute to the operational status of the associated virtual-address. The default is \'yes\'.',
      default: 'yes',
    },
    auto_lasthop: {
      type: 'Optional[Enum["default", "enabled", "disabled"]]',
      desc: 'Specifies whether to automatically map last hop for pools or not. The default is to use next level\'s default.',
      default: 'default',
    },
    bwc_policy: {
      type: 'Optional[String]',
      desc: 'Specifies the name of an existing bwc policy for the virtual server to use to enforce a throughput policy for incoming network traffic.',
    },
    cmp_enabled: {
      type: 'Optional[Enum["yes", "no"]]',
      desc: '	Enables or disables clustered multi-processor (CMP) acceleration. This feature applies to certain platforms only. The default value is enabled.',
      default: 'no',
    },
    connection_limit: {
      type: 'Optional[Integer]',
      desc: 'Specifies the maximum number of concurrent connections that the system allows for this virtual server. A value of 0 specifies that there is no limit. The default is 0.',
    },
    destination: {
      type: 'Optional[String]',
      desc: 'Specifies the <virtual-address>:<port> on which the virtual server listens for connections. <virtual-address> can be an IP address or a previously created virtual-address.',
    },
    dhcp_relay: {
      type: 'Optional[Boolean]',
      desc: '	Specifies a virtual server that relays all received DHCP requests to all pool members. If there is no pool, the received request get dropped.',
    },
    disabled: {
      type: 'Optional[Boolean]',
      desc: 'Note that when you disable a virtual server, the virtual server no longer accepts new connection requests. However, it allows current connections to finish processing before going to a down state.',
    },
    enabled: {
      type: 'Optional[Boolean]',
      desc: 'Specifies the state of the virtual server. The default value is enabled. This brings the virtual server to a state where it will accept new connections.',
    },
    fallback_persistence: {
      type: 'Optional[String]',
      desc: 'Specifies a fallback persistence profile for the virtual server to use when the default persistence profile is not available.',
    },
    flow_eviction_policy: {
      type: 'Optional[String]',
      desc: 'Specifies the flow eviction policy for the virtual server to use when the connection limit is approached',
    },
    fw_enforced_policy: {
      type: 'Optional[String]',
      desc: 'Enforced firewall policy.',
    },
    fw_staged_policy: {
      type: 'Optional[String]',
      desc: 'Staged firewall policy.',
    },
    gtm_score: {
      type: 'Optional[Integer]',
      desc: 'Specifies a score that is associated with the virtual server. Global Traffic Manager (GTM) can rely on this value to load balance traffic in a proportional manner.',
    },
    internal: {
      type: 'Optional[Boolean]',
      desc: 'Specifies an internal virtual server that handles requests on the side for a parent virtual server, such as content adaptation.',
    },
    ip_forward: {
      type: 'Optional[Boolean]',
      desc: 'Specifies a virtual server like other virtual servers, except that the virtual server has no pool members to load balance. The virtual server simply forwards the packet directly to the destination IP address specified in the client request. Internal virtual servers do not receive external connections, instead they are specified by name by profiles in the parent virtual server (see ltm profile request-adapt and ltm profile response-adapt). Since internal virtual servers do not listen for external connections, not all attributes are used for internal virtual servers. The destination, mask, translate-address, translate-port, vlans, vlans-disabled and vlans-enabled attributes are set by the system and any attempt to change them will have no effect.',
    },
    ip_intelligence_policy: {
      type: 'Optional[String]',
      desc: 'Name of the IP Intelligence (Dynamic White/Black List) policy that is attached.',
    },
    ip_protocol: {
      type: 'Optional[Enum["ah", "any", "bna", "esp", "etherip", "gre", "icmp", "ipencap", "ipv6", "ipv6-auth", "ipv6-crypt", "ipv6-icmp", "isp-ip", "mux", "ospf", "sctp", "tcp", "udp", "udplite"]]',
      desc: 'Specifies the IP protocol for which you want the virtual server to direct traffic. Sample protocol names are tcp and udp.',
    },
    l2_forward: {
      type: 'Optional[Boolean]',
      desc: 'Specifies a virtual server that shares the same IP address as a node in an associated VLAN. This is the type of virtual server you create when you want to create a VLAN group.',
    },
    last_hop_pool: {
      type: 'Optional[String]',
      desc: 'Specifies the name of the last hop pool that you want the virtual server to use to direct reply traffic to the last hop router.',
    },
    mask: {
      type: 'Optional[String]',
      desc: 'Specifies the netmask for a network virtual server only. This setting is required for a network virtual server. The netmask clarifies whether the host bit is an actual zero or a wildcard representation.',
      default: '255.255.255.255',
    },
    mirror: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Enables or disables state mirroring. You can use state mirroring to maintain the same state information in the standby unit that is in the active unit, allowing transactions such as FTP file transfers to continue as though uninterrupted. The default value is disabled.',
      default: 'disabled',
    },
    mobile_app_tunnel: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: '',
      default: 'disabled',
    },
    nat64: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Specifies whether this virtual does NAT64 translation',
      default: 'disabled',
    },
    per_flow_request_access_policy: {
      type: 'Optional[String]',
      desc: '',
    },
    pool: {
      type: 'Optional[String]',
      desc: 'Specifies a default pool to which you want the virtual server to automatically direct traffic.',
    },
    rate_class: {
      type: 'Optional[String]',
      desc: 'Specifies the name of an existing rate class for the virtual server to use to enforce a throughput policy for incoming network traffic.',
    },
    rate_limit: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Specifies the maximum number of connections per second allowed for a virtual server. The default value is \'disabled\'.',
      default: 'disabled',
    },
    rate_limit_dst_mask: {
      type: 'Optional[Integer]',
      desc: 'Specifies a mask, in bits, to be applied to the destination address as part of the rate limiting. The default value is ‘0’, which is equivalent to using the entire address - ‘32’ in IPv4, or ‘128’ in IPv6.',
      default: 0,
    },
    rate_limit_mode: {
      type: 'Optional[Enum["destination", "object-destination", "object-source-destination", "source-destination", "object", "object-source", "source"]]',
      desc: '	Indicates whether the rate limit is applied per virtual object, per source address, per destination address, or some combination thereof. The default value is ‘object’, which does not use the source or destination address as part of the key',
      default: 'object',
    },
    rate_limit_src_mask: {
      type: 'Optional[Integer]',
      desc: 'Specifies a mask, in bits, to be applied to the source address as part of the rate limiting. The default value is ‘0’, which is equivalent to using the entire address - ‘32’ in IPv4, or ‘128’ in IPv6.',
      default: 0,
    },
    reject: {
      type: 'Optional[Boolean]',
      desc: 'Specifies that the BIG-IP system rejects any traffic destined for the virtual server IP address.',
    },
    service_down_immediate_action: {
      type: 'Optional[Enum["none", "reset", "drop"]]',
      desc: 'Specifies the immediate action to take upon the receipt of the initial client’s SYN packet if the availability status of the virtual server is Offline or Unavailable. This is supported for the virtual server of Standard type and TCP protocol. The default value is none.',
      default: 'none',
    },
    saervice_policy: {
      type: 'Optional[String]',
      desc: 'Name of the Service Policy.',
    },
    source: {
      type: 'Optional[String]',
      desc: 'Specifies an IP address or network from which the virtual server will accept traffic.',
      default: '0.0.0.0/0'
    },
    source_port: {
      type: 'Optional[Enum["preserve", "preserve-strict", "change"]]',
      desc: 'Specifies whether the system preserves the source port of the connection. The default is preserve. Use of the preserve-strict setting should be restricted to UDP only under very special circumstances such as nPath or transparent (that is, no translation of any other L3/L4 field), where there is a 1:1 relationship between virtual IP addresses and node addresses, or when clustered multi-processing (CMP) is disabled. The change setting is useful for obfuscating internal network addresses.',
      default: 'preserve',
    },
    stateless: {
      type: 'Optional[Boolean]',
      desc: 'The virtual server load balances each packet to a pool member, but no connection state is maintained.',
    },
    throughput_capacity: {
      type: 'Optional[Variant[Integer,String]]',
      desc: 'Specifies the max throughput capacity of this virtual server in Mbps.',
    },
    translate_address: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Enables or disables address translation for the virtual server. Turn address translation off for a virtual server if you want to use the virtual server to load balance connections to any address. This option is useful when the system is load balancing devices that have the same IP address.',
      default: 'enabled',
    },
    translate_port: {
      type: 'Optional[Enum["enabled", "disabled"]]',
      desc: 'Enables or disables port translation. Turn port translation off for a virtual server if you want to use the virtual server to load balance connections to any service.',
      default: 'enabled',
    },
    transparent_next_hop: {
      type: 'Optional[String]',
      desc: 'Specifies the egress interface for traffic and enables layer 2 (MAC) address preservation. Layer 2 address preservation disables layer 3 (IP/IPv6) address translation.',
    },
    vlans_disabled: {
      type: 'Optional[Boolean]',
      desc: 'Disables the virtual server on the VLANs specified by the VLANs option.',
    },
    vlans_enabled: {
      type: 'Optional[Boolean]',
      desc: 'Enables the virtual server on the VLANs specified by the VLANs option.',
    },
    vs_index: {
      type: 'Optional[Integer]',
      desc: 'Specifies the index of the virtual server. The index is used to determine the order in which the virtual servers are evaluated when a request arrives.',
      default: 0,
      behaviour: :read_only,
    },
    source_address_translation: {
      type: <<-EOS,
        Optional[Struct[{
          Optional['type'] => Enum['automap', 'snat', 'none'],
          Optional['pool'] => String,
        }]]
      EOS
      desc: 'Specifies the type of source address translation associated with the specified virtual server.',
      struct_skip_keys: [:poolReference],
      default: {
        'type' => 'none',
      }
    }
  },
}))
